import re

from nerwhal.recognizer_bases.re_recognizer import ReRecognizer


class EmailRecognizer(ReRecognizer):
    """Recognize email addresses in the most common formats.

    This recognizer does not aim at exactly implementing the email syntax specification (which is rather complex), but at
    finding a superset of the specification. Thus, it may return results that aren't actually compliant/valid email addresses.

    Also, some addresses that use a less common syntax might be missed. Address syntax features that aren't yet supported
    include but are not limited to:
    - IP address domains
    - domain names without TLD
    - UTF8 characters in local and domain part
    - special characters inside quoted strings
    """

    TAG = "EMAIL"
    SCORE = 0.95
    FLAGS = re.VERBOSE
    CONTEXT_WORDS = ["E-Mail", "email", "Mail", "mail"]

    @property
    def regexp(self):
        return r"""[a-zA-Z0-9!#$%&'"*+\-/=?^_`{|}~]  # the first character cannot be a dot
                   [a-zA-Z0-9!#$%&'"*+\-/=?^_`{|}~\.]*  # local part
                   @
                   ((?!-)[a-zA-Z0-9\-]{0,62}[a-zA-Z0-9]\.)+[a-zA-Z]{2,63}  # domain part"""



import pytest

from nerwhal.integrated_recognizers.money_recognizer import MoneyRecognizer


@pytest.fixture(scope="module")
def de_backend(setup_backend):
    recognizer = MoneyRecognizer
    backend = setup_backend(recognizer.BACKEND, language="de")
    backend.register_recognizer(recognizer)
    return backend


@pytest.fixture(scope="module")
def en_backend(setup_backend):
    recognizer = MoneyRecognizer
    backend = setup_backend(recognizer.BACKEND, language="en")
    backend.register_recognizer(recognizer)
    return backend


def test_de_euro(de_backend, embed):
    text = "Das Brot kostet 1,50 €."
    ents = de_backend.run(text)
    assert embed(text, ents) == "Das Brot kostet MONEY."


def test_de_euro_w_dots(de_backend, embed):
    text = "Er möchte 1.000.000 € im Lotto gewinnen."
    ents = de_backend.run(text)
    assert embed(text, ents) == "Er möchte MONEY im Lotto gewinnen."


def test_de_dollar(de_backend, embed):
    text = "Das Brot kostet 1,50 $."
    ents = de_backend.run(text)
    assert embed(text, ents) == "Das Brot kostet MONEY."


def test_us_dollar(en_backend, embed):
    text = "The bread costs $1.50."
    ents = en_backend.run(text)
    assert embed(text, ents) == "The bread costs MONEY."


def test_us_euro_w_dots_and_space(en_backend, embed):
    text = "He'd like to win € 1,000,000 in the lottery."
    ents = en_backend.run(text)
    assert embed(text, ents) == "He'd like to win MONEY in the lottery."


def test_us_euro(en_backend, embed):
    text = "The bread costs €1.50."
    ents = en_backend.run(text)
    assert embed(text, ents) == "The bread costs MONEY."



import os
from pathlib import Path


def list_all():
    all_examples = []
    integrated_recognizers_directory = Path(__file__).parent
    for root, _, files in os.walk(integrated_recognizers_directory):
        for file in files:
            if file.endswith("_recognizer.py"):
                all_examples.append(os.path.join(root, file))
    return all_examples



def combine(ents, strategy=None):
    """Process a list of entities using different strategies to deal with overlapping or conflicting entities.

    This function offers the following strategies:
    - None: Keep all entities.
    - `disjunctive_union`: This function keeps and returns all entities, but assumes their character spans to be disjunct;
        an error is thrown if two entities overlap.
    - `fusion`: Resolve overlapping entities of different tags by keeping that with the highest score.
    - `smart-fusion`: Like fusion, but add scores if one entity occurs multiple times with the same tag.
    """
    ents.sort(key=lambda ent: (ent.start_char, ent.end_char, 1.0 - ent.score, ent.tag))

    if strategy is None:
        return ents

    if strategy == "disjunctive_union":
        combined = _disjunctive_union_strategy(ents)
    elif strategy == "fusion":
        combined = _fusion_strategy(ents)
    elif strategy == "smart-fusion":
        combined = _smart_fusion_strategy(ents)
    else:
        raise ValueError(f"Unknown aggregation strategy {strategy}")

    return combined


def _disjunctive_union_strategy(ents):
    """This strategy ensures that all entities are disjoint.

    Checks that all entities are disjoint by comparing the end of the previous entity with the start of the current one.
    """
    prev_ent_end = 0
    for ent in ents:
        if prev_ent_end > ent.start_char:
            raise AssertionError(
                f"All entities were assumed to be disjunct, but {ent.text} ({ent.start_char}-{ent.end_char}) wasn't"
            )

        prev_ent_end = ent.end_char
    return ents


def _overlapping(first, second):
    return first.start_char <= second.start_char < first.end_char


def _fusion_strategy(ents):
    """This strategy resolves overlapping named entities by giving those with higher scores priority."""
    res = []
    prev_ent = None
    for idx, ent in enumerate(ents):
        if prev_ent is None:
            res.append(ent)
            prev_ent = ent
            continue

        if not _overlapping(prev_ent, ent):
            res.append(ent)
            prev_ent = ent
            continue

        if _overlapping(prev_ent, ent):
            if prev_ent.score < ent.score:
                res[-1] = ent
                prev_ent = ent
                continue
            elif prev_ent.score > ent.score:
                continue
            else:
                prev_ent_length = prev_ent.end_char - prev_ent.start_char
                ent_length = ent.end_char - ent.start_char
                if prev_ent_length < ent_length:
                    res[-1] = ent
                    prev_ent = ent
                    continue
                else:
                    continue

    return res


def _is_same(ent_a, ent_b):
    return ent_a.start_char == ent_b.start_char and ent_a.end_char == ent_b.end_char and ent_a.tag == ent_b.tag


def _smart_fusion_strategy(ents):
    """This strategy adds the entities scores if one entity occurs multiple times with the same tag, and afterwards
    applies the fusion strategy."""
    deduplicated_ents = []
    for idx, ent in enumerate(ents):
        next_ent = ents[idx + 1] if idx + 1 < len(ents) else None

        if next_ent and _is_same(next_ent, ent):
            next_ent.score = min(next_ent.score + ent.score, 1.0)
            continue

        deduplicated_ents.append(ent)

    return _fusion_strategy(deduplicated_ents)



from typing import Type

from spacy.pipeline import EntityRuler

from .base import Backend
from nerwhal.recognizer_bases import EntityRulerRecognizer
from nerwhal.types import NamedEntity
from nerwhal.nlp_utils import (
    configure_spacy_entity_extension_attributes,
    set_spacy_entity_extension_attributes,
    load_spacy_nlp,
)

configure_spacy_entity_extension_attributes()


class EntityRulerBackend(Backend):
    """This backend recognizes entities using the spaCy EntityRuler.

    See https://spacy.io/usage/rule-based-matching#entityruler for more information about the EntityRuler.
    """

    def __init__(self, language):
        self.nlp = load_spacy_nlp(language, disable_components=["tagger", "parser", "ner"])

    def register_recognizer(self, recognizer_cls: Type[EntityRulerRecognizer]):
        recognizer = recognizer_cls()

        recognizer_name = recognizer_cls.__name__
        ruler = EntityRuler(self.nlp)
        self.nlp.add_pipe(ruler, recognizer_name)
        rules = [{"label": recognizer.TAG, "pattern": pattern} for pattern in recognizer.patterns]
        ruler.add_patterns(rules)
        self.nlp.add_pipe(
            set_spacy_entity_extension_attributes(recognizer.SCORE, recognizer_name),
            name="label_" + recognizer_name,
            after=recognizer_name,
        )

    def run(self, text):
        doc = self.nlp(text)

        ents = []
        for ent in doc.ents:
            ents += [NamedEntity(ent.start_char, ent.end_char, ent.label_, ent.text, ent._.score, ent._.recognizer)]

        return ents



from typing import List

from nerwhal.nlp_utils import load_spacy_nlp, configure_spacy_entity_extension_attributes
from nerwhal.types import Token

configure_spacy_entity_extension_attributes()


class Tokenizer:
    """Tokenize text.

    Use the Tokenizer by first tokenizing the text, and then calling the getter methods.
    """

    def __init__(self, language):
        self.nlp = load_spacy_nlp(language, disable_components=["tagger", "ner"])
        self.doc = None

    def tokenize(self, text):
        self.doc = self.nlp(text)

    def get_tokens(self):
        return self._to_nerwhal_tokens(self.doc)

    def get_sentence_for_token(self, idx, exclude_tokens: List[int] = None):
        """Return the sentence that contains the token with given index.

        :param idx: the index of the token in the text
        :param exclude_tokens: exclude the tokens with the given indices from the returned sentence
        """
        spacy_tokens = self.doc[idx].sent

        if exclude_tokens:
            spacy_tokens = [token for token in spacy_tokens if token.i not in exclude_tokens]

        return self._to_nerwhal_tokens(spacy_tokens)

    def _to_nerwhal_tokens(self, spacy_tokens):
        """Translates the spaCy to the NERwhal token representation."""
        return [
            Token(
                text=token.text,
                has_ws=token.whitespace_ == " ",
                br_count=token.text.count("\n"),
                start_char=token.idx,
                end_char=token.idx + len(token),
            )
            for token in spacy_tokens
        ]



from abc import abstractmethod

from .base import Recognizer


class EntityRulerRecognizer(Recognizer):
    """The base class for all spaCy EntityRuler based recognizers."""

    BACKEND = "entity-ruler"

    @property
    @abstractmethod
    def patterns(self):
        """This property returns the patterns for the entity ruler.

        See https://spacy.io/usage/rule-based-matching#entityruler for more information on the patterns and
        https://spacy.io/api/token#attributes for the available token attributes.
        """
        pass



from abc import ABC, abstractmethod


class Recognizer(ABC):
    """This interface is the base class for all recognizers.

    Recognizers are basically configurations for the backends. Each backend runs one or more recognizers to search for
    entities.

    For efficiency reasons do the heavy lifting of loading files or expensive computations in the `__init__` method instead of
    in the properties. The recognizer is only initialized once, but the properties are called on every run.
    """

    CONTEXT_WORDS = []

    @property
    @abstractmethod
    def BACKEND(self):
        """The backend that this recognizer operates with."""
        pass

    @property
    @abstractmethod
    def TAG(self):
        """The recognizer finds named entities for this tag."""
        pass

    @property
    @abstractmethod
    def SCORE(self):
        """The confidence score for a result of this recognizer being a named entity for its tag, i.e. the precision of this
        recognizer."""
        pass



import pytest

from nerwhal.integrated_recognizers.email_recognizer import EmailRecognizer


@pytest.fixture
def backend(setup_backend):
    recognizer = EmailRecognizer
    backend = setup_backend(recognizer.BACKEND)
    backend.register_recognizer(recognizer)
    return backend


def test_simple_email(backend, embed):
    text = "Meine E-Mail Adresse ist obiwan@gmail.com."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_dotted_email(backend, embed):
    text = "Meine E-Mail Adresse ist han.solo@example.com."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_disposable_style_with_plus(backend, embed):
    text = "Meine E-Mail Adresse ist disposable.style.with+symbol@example.com und user.name+tag+sorting@example.com."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL und EMAIL."


def test_with_hyphen(backend, embed):
    text = "Meine E-Mail Adresse ist mail-with-hyphen@example.com."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_one_letter_local_part(backend, embed):
    text = "Meine E-Mail Adresse ist x@example.com."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_special_chars_in_local_part(backend, embed):
    text = "Meine E-Mail Adresse ist mailhost!user%example.com@example.org."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_host_with_hyphen(backend, embed):
    text = "Meine E-Mail Adresse ist test@my-host.com."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_one_letter_host(backend, embed):
    text = "Meine E-Mail Adresse ist test@x.com."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_host_with_many_labels(backend, embed):
    text = "Meine E-Mail Adresse ist test@de.example.uk."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_host_with_special_tld(backend, embed):
    text = "Meine E-Mail Adresse ist test@example.foundation."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine E-Mail Adresse ist EMAIL."


def test_no_at_character(backend, embed):
    text = "Meine E-Mail Adresse ist test.example.com."
    ents = backend.run(text)
    assert len(ents) == 0


def test_cannot_start_with_dot(backend, embed):
    text = "...mail@example.com..."
    ents = backend.run(text)
    assert embed(text, ents) == "...EMAIL..."



import pytest

from nerwhal.integrated_recognizers.de.de_date_recognizer import DeDateRecognizer


@pytest.fixture(scope="module")
def backend(setup_backend):
    recognizer = DeDateRecognizer
    backend = setup_backend(recognizer.BACKEND, language="de")
    backend.register_recognizer(recognizer)
    return backend


# DIN 1355-1


def test_current_1355_1(backend, embed):
    text = "Der 25.06.1999 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_future_1355_1(backend, embed):
    text = "Der 25.12.2978 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_historic_1355_1(backend, embed):
    text = "Der 25.12.768 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_abbreviated_1355_1(backend, embed):
    text = "Der 5.12.74 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_1355_1_invalid_month(backend, embed):
    text = "Der 24.13.1999 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert len(ents) == 0


def test_1355_1_invalid_day(backend, embed):
    text = "Der 35.12.1999 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert len(ents) == 0


# DIN 5008


def test_current_5008(backend, embed):
    text = "Der 1999-06-25 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_future_5008(backend, embed):
    text = "Der 2978-12-25 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_5008_invalid_month(backend, embed):
    text = "Der 1999-13-25 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert len(ents) == 0


def test_5008_invalid_day(backend, embed):
    text = "Der 1999-12-35 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert len(ents) == 0


# Written out


def test_current_written_out(backend, embed):
    text = "Der 25. Juni 1999 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_future_written_out(backend, embed):
    text = "Der 25. Dezember 2978 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_historic_written_out(backend, embed):
    text = "Der 25. Dezember 768 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_abbreviated_written_out(backend, embed):
    text = "Der 5. Dez. 74 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "Der DATE ist ein zufälliges Datum."


def test_with_day_written_out(backend, embed):
    text = "Montag, 5. Dez. 74 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert embed(text, ents) == "DATE ist ein zufälliges Datum."


def test_written_out_invalid_month(backend, embed):
    text = "Der 24. Foobar 1999 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert len(ents) == 0


def test_written_out_invalid_day(backend, embed):
    text = "Der 35. Dezember 1999 ist ein zufälliges Datum."
    ents = backend.run(text)
    assert len(ents) == 0



import re

from nerwhal.recognizer_bases.re_recognizer import ReRecognizer


class PhoneNumberRecognizer(ReRecognizer):
    """Recognize phone numbers that follow several common German national and international patterns.

    This recognizer aims at finding numbers of DIN 5008 and E.123 style, Microsoft's canonical number format, and some
    non-standard but frequently used styles. Because of the freedom that a user has in writing a phone number and splitting it
    into differently sized blocks, it is difficult to find all occurrences and to not produce too many false positives.
    You can refer to the regular expression below or the tests to see what sequences are detected.

    Consider the following remarks and limitations:
    - as there are many ways in which a phone number can be formatted, we cannot match all of them
    - making the patterns too universal may produce false positives for other sequences of digits such as credit card numbers
    - choosing the quantifiers for the digits in a black constitutes a trade-off between detecting more phone numbers in
      less common formats vs producing false positives
    - especially, numbers in foreign styles or numbers in German style that represent a foreign number aren't tested and may
      not be recognized
    """

    TAG = "PHONE"
    SCORE = 0.8
    GROUP = 1
    FLAGS = re.VERBOSE
    CONTEXT_WORDS = ["number", "Nummer", "tel.", "Tel.", "mobil", "Telefon", "telephone"]

    COUNTRY_CODE_REGEX = r"\(?(\+|00)\)?\d{1,3}"
    AREA_CODE = r"(\(?0\)?[ ]?\d{1,4}|\(?\d{1,5}\)?)"
    BASE_NUMBER = r"\d{1,4}[ -]?\d{1,4}[ -]?\d{1,2}"
    EXTENSION = r"([ ]?-[ ]?\d{1,3}|[ ]\d{0,3}[ ]?\/[ ]?\d{1,3})"

    @property
    def regexp(self):
        return rf"""# there should not be a digit in the preceding two characters and no dot directly before a number
                    (?:^|^[^\.\d]|[^\d][^\.\d])
                    (  # the number will be in the matching group 1
                    # >>> International numbers
                    # Microsoft's canonical format, e.g. +49 (030) 12345, +49 (0) 30 12345
                    # E.123 style with blocks, e.g. +22 607 123 4567
                    # Abbreviated, e.g. tel:+49-30-1234567
                    ({self.COUNTRY_CODE_REGEX}[ \.\-]?)?
                    # >>> National numbers
                    # DIN 5008, e.g. 030 12345-67
                    # E.123, e.g. (030) 12345 0 / 67
                    # Non-standard format, e.g. 0 30 / 12 34 56
                    {self.AREA_CODE}([ -]|([ ]?\/[ ]?))?{self.BASE_NUMBER}{self.EXTENSION}?
                    )
                    # the following two characters should not be digits
                    (?:$|[^\d]$|[^\d][^\d])
                    """



from abc import abstractmethod

from .base import Recognizer


class ReRecognizer(Recognizer):
    """The base class for all regular expression based recognizers."""

    BACKEND = "re"
    GROUP = 0  # the name or number of the capture group, if it is 0 the whole match will be taken
    FLAGS = 0  # re compiler flags, e.g. re.VERBOSE

    @property
    @abstractmethod
    def regexp(self):
        """A string that describes the regular expression pattern for this recognizer."""
        pass



from importlib import import_module


def load(backend):
    """Dynamically load the requested backend module."""

    if backend == "re":
        mod = ".re_backend"
        cls = "ReBackend"
    elif backend == "flashtext":
        mod = ".flashtext_backend"
        cls = "FlashtextBackend"
    elif backend == "entity-ruler":
        mod = ".entity_ruler_backend"
        cls = "EntityRulerBackend"
    else:
        raise ValueError(f"Unknown backend type {backend}")

    backend_cls = getattr(import_module(mod, package="nerwhal.backends"), cls)
    return backend_cls



import pytest

from nerwhal.integrated_recognizers.phone_number_recognizer import PhoneNumberRecognizer


@pytest.fixture(scope="module")
def backend(setup_backend):
    recognizer = PhoneNumberRecognizer
    backend = setup_backend(recognizer.BACKEND)
    backend.register_recognizer(recognizer)
    return backend


# DIN 5008


def test_international_din_5008(backend, embed):
    text = "Meine Telefonnummer ist +49 30 12345-67."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_national_din_5008(backend, embed):
    text = "Meine Telefonnummer ist 01234 5678."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_national_din_5008_hyphenated_extension(backend, embed):
    text = "Meine Telefonnummer ist 030 12345-67."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_national_din_5008_value_added_service_with_code(backend, embed):
    text = "Meine Telefonnummer ist 0900 5 123456."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


# Microsoft's canonical format


def test_microsofts_canonical_format(backend, embed):
    text = "Meine Telefonnummer ist +49 (30) 123456."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_microsofts_canonical_format_with_extension(backend, embed):
    text = "Meine Telefonnummer ist +49 (30) 12345 - 67."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_microsofts_canonical_format_with_extension_no_space(backend, embed):
    text = "Meine Telefonnummer ist +49 (30) 123456-78."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


# E.123


def test_international_e_123(backend, embed):
    text = "Meine Telefonnummer ist +49 89 123 456 78."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_international_e_123_company_center(backend, embed):
    text = "Meine Telefonnummer ist +49 89 123 456 0."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_national_e_123(backend, embed):
    text = "Meine Telefonnummer ist (042) 123 4567."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_national_e_123_with_options(backend, embed):
    text = "Meine Telefonnummer ist (030) 12345 0 / 67."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


# Others


def test_mobile_number(backend, embed):
    text = "Meine Telefonnummer ist 0160 1234567."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_freephone(backend, embed):
    text = "Meine Telefonnummer ist 0800 123456."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_freephone_without_zero(backend, embed):
    text = "Meine Telefonnummer ist 800 123456."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


# Non-standard


def test_double_zero_international(backend, embed):
    text = "Meine Telefonnummer ist 0049 1234 56789."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_number_abbreviated(backend, embed):
    text = "Sie können uns erreichen unter tel.:+49-30-1234567."
    ents = backend.run(text)
    assert embed(text, ents) == "Sie können uns erreichen unter tel.:PHONE."


def test_dotted_format(backend, embed):
    text = "Meine Telefonnummer ist +49.3012345."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_national_not_standardized(backend, embed):
    text = "Meine Telefonnummer ist 0 30 / 12 34 56."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


def test_international_not_standardized_optional_zero(backend, embed):
    text = "Meine Telefonnummer ist +49 (0) 30 12345-67."
    ents = backend.run(text)
    assert embed(text, ents) == "Meine Telefonnummer ist PHONE."


# US style


def test_local(backend, embed):
    text = "My number is 234-1234."
    ents = backend.run(text)
    assert embed(text, ents) == "My number is PHONE."


def test_domestic(backend, embed):
    text = "My number is (543) 234-1234."
    ents = backend.run(text)
    assert embed(text, ents) == "My number is PHONE."


def test_international_from_us(backend, embed):
    text = "My number is +1-543-234-1234."
    ents = backend.run(text)
    assert embed(text, ents) == "My number is PHONE."


def test_international_from_abroad(backend, embed):
    text = "My number is 001-543-234-1234."
    ents = backend.run(text)
    assert embed(text, ents) == "My number is PHONE."


# Not phone numbers


def test_credit_card(backend):
    # The matcher pattern that matches most tokens takes priority. So a credit card matcher will take priority over
    # two number matches in 1234 1234 1234 1234
    text = "Meine Kreditkarten-Nummer ist 1234123412341234, anders geschrieben 1234 1234 1234 1234."
    ents = backend.run(text)
    assert len(ents) == 0


def test_date(backend):
    text = "Heute ist der 12.12.2012."
    ents = backend.run(text)
    assert len(ents) == 0


def test_us_date(backend):
    text = "Today is the 01/02/2000."
    ents = backend.run(text)
    assert len(ents) == 0



from typing import Type

from flashtext import KeywordProcessor

from .base import Backend
from nerwhal.types import NamedEntity
from ..recognizer_bases import FlashtextRecognizer


class FlashTextBackend(Backend):
    """This backend recognizes entities using the FlashText algorithm.

    See https://flashtext.readthedocs.io/en/latest/ for more information about the FlashText algorithm.
    """

    def __init__(self):
        self.keyword_processor = KeywordProcessor()

    def register_recognizer(self, recognizer_cls: Type[FlashtextRecognizer]):
        recognizer = recognizer_cls()

        key = f"{recognizer.TAG}:{recognizer.SCORE}:{recognizer_cls.__name__}"
        keyword_dict = {key: recognizer.keywords}
        self.keyword_processor.add_keywords_from_dict(keyword_dict)

    def run(self, text):
        keywords = self.keyword_processor.extract_keywords(text, span_info=True)

        ents = []
        for keyword_key, start, end in keywords:
            tag, score, recognizer_name = keyword_key.split(":")
            ent = NamedEntity(start, end, tag, text[start:end], float(score), recognizer_name)
            ents.append(ent)
        return ents



from abc import abstractmethod

from .base import Recognizer


class FlashtextRecognizer(Recognizer):
    """The base class for all FlashText based recognizers."""

    BACKEND = "flashtext"

    @property
    @abstractmethod
    def keywords(self):
        """The list of keywords that the FlashText backend is searching for."""
        pass



import pytest

from nerwhal.backends.stanza_ner_backend import StanzaNerBackend


@pytest.fixture(scope="module")
def stanza_ner_backend():
    return StanzaNerBackend("de")


@pytest.mark.stanza
def test_simple_person(stanza_ner_backend, embed):
    text = "Yoda war mein Lehrmeister."
    ents = stanza_ner_backend.run(text)
    assert embed(text, ents) == "PER war mein Lehrmeister."


@pytest.mark.stanza
def test_simple_location(stanza_ner_backend, embed):
    text = "Wer kennt das schöne Deutschland nicht?"
    ents = stanza_ner_backend.run(text)
    assert embed(text, ents) == "Wer kennt das schöne LOC nicht?"


@pytest.mark.stanza
def test_multiple_ents(stanza_ner_backend, embed):
    text = "Meister Yoda macht in Berlin Urlaub."
    ents = stanza_ner_backend.run(text)
    assert embed(text, ents) == "Meister PER macht in LOC Urlaub."
